# -*- coding: utf-8 -*-
"""
copyright:
    The MagPy Development Team
"""

# import order matters - NamedTemporaryFile must be one of the first!
__all__ = ['core','lib','app'
]

#import martas.core
import martas.app
#import magpy.core

#from .version import __version__
