:root {
	  --background-color: #fff;
	  --text-color: black;
	  --menu-color: #888888;
	  --wrap-bgcolor: #fff;
	  --wrap-color: #EDE9E8;
	  --box-bgcolor: #EDE9E8;
	  --box-color: #444;
	  --shadow-color: #999;
	  --highlight: #cdcdcd;
	  --primary: hsl(12, 70%, 40%);
	}

.navbar-light .navbar-brand {
  color: var(--menu-color);
}

body {
	background-color: var(--background-color);
	margin: 10px;
	font-family: sans-serif;
}

.mwrap {
	display: grid;
	grid-gap: 10px;
	grid-template-columns: 1fr 1fr 1fr 1fr 1fr 1fr 1fr 1fr 1fr;
	background-color: var(--wrap-bgcolor);
	color: var(--wrap-color);
	}
.mbox {
	background-color: #EDE9E8;
	color: #444;
	border-radius: 5px;
	padding: 10px;
	font-size: 90%;
	}
.ma {
	grid-column: 1 / 3;
	grid-row: 1;
	}
.mb {
	grid-column: 3 / 4 ;
	grid-row: 1 ;
	}
.mc {
	grid-column: 4 / 5 ;
	grid-row: 1 ;
	}
.md {
	grid-column: 5 / 6 ;
	grid-row: 1 ;
	}
.me {
	grid-column: 6 / 7 ;
	grid-row: 1 ;
	}
.mf {
	grid-column: 7 / 8 ;
	grid-row: 1 ;
	}
.mg {
	grid-column: 8 / 9 ;
	grid-row: 1 ;
	}
.mh {
	grid-column: 9 / 10 ;
	grid-row: 1 ;
	}
.mi {
	grid-column: 1 / 4 ;
	grid-row: 2 ;
	}
.mj {
	grid-column: 1 / 2 ;
	grid-row: 3 ;
	}
.mk {
	grid-column: 2 / 3 ;
	grid-row: 3 ;
	}
.ml {
	grid-column: 3 / 4 ;
	grid-row: 3 ;
	}
.mm {
	grid-column: 1 / 2 ;
	grid-row: 4 ;
	}
.mn {
	grid-column: 2 / 3 ;
	grid-row: 4 ;
	}
.mo {
	grid-column: 3 / 4 ;
	grid-row: 4 ;
	}
.mp {
	grid-column: 4 / 10 ;
	grid-row: 2 / 5 ;
	}
.mq {
	grid-column: 1 / 4 ;
	grid-row: 5 / 7 ;
	}
.mr {
	grid-column: 4 / 10 ;
	grid-row: 5 ;
	}
.mt {
	grid-column: 4 / 7 ;
	grid-row: 6 ;
	}
.mu {
	grid-column: 7 / 10 ;
	grid-row: 6 ;
	}

@media only screen and (max-width: 780px) and (min-width: 480px){
	body {
		margin: 2px;
	}
	.mwrap { grid-template-columns: 1fr 1fr 1fr }
	.mbox {
	border-radius: 2px;
	padding: 5px;
	font-size: 90%;
	}

.ma {
	grid-column: 1 / 3;
	grid-row: 1;
	}
.mb {
	grid-column: 3 / 4 ;
	grid-row: 1 ;
	}
.mc {
	grid-column: 1 / 2 ;
	grid-row: 2 ;
	}
.md {
	grid-column: 2 / 3 ;
	grid-row: 2 ;
	}
.me {
	grid-column: 3 / 4 ;
	grid-row: 2 ;
	}
.mf {
	grid-column: 1 / 2 ;
	grid-row: 3 ;
	}
.mg {
	grid-column: 2 / 3 ;
	grid-row: 3 ;
	}
.mh {
	grid-column: 3 / 4 ;
	grid-row: 3 ;
	}
.mi {
	grid-column: 1 / 4 ;
	grid-row: 4 ;
	}
.mj {
	grid-column: 1 / 2 ;
	grid-row: 6 ;
	}
.mk {
	grid-column: 2 / 3 ;
	grid-row: 6 ;
	}
.ml {
	grid-column: 3 / 4 ;
	grid-row: 6 ;
	}
.mm {
	grid-column: 1 / 2 ;
	grid-row: 7 ;
	}
.mn {
	grid-column: 2 / 3 ;
	grid-row: 7 ;
	}
.mo {
	grid-column: 3 / 4 ;
	grid-row: 7 ;
	}
.mp {
	grid-column: 1 / 4 ;
	grid-row: 5 ;
	}
.mq {
	grid-column: 1 / 4 ;
	grid-row: 8;
	}
.mr {
	grid-column: 1 / 4 ;
	grid-row: 9 ;
	}
.mt {
	grid-column: 1 / 4 ;
	grid-row: 10 ;
	}
.mu {
	grid-column: 1 / 4 ;
	grid-row: 11 ;
	}
}

@media only screen and (max-width: 480px) {
	body {
		margin: 10px;
	}
	.mwrap { grid-template-columns: 1fr 1fr }
	.mbox {
	border-radius: 1px;
	padding: 5px;
	font-size: 90%;
	}
.ma {
	grid-column: 1 / 3;
	grid-row: 1;
	}
.mb {
	grid-column: 1 / 2 ;
	grid-row: 2 ;
	}
.mc {
	grid-column: 2 / 3 ;
	grid-row: 2 ;
	}
.md {
	grid-column: 1 / 2 ;
	grid-row: 3 ;
	}
.me {
	grid-column: 2 / 3 ;
	grid-row: 3 ;
	}
.mf {
	grid-column: 1 / 2 ;
	grid-row: 4 ;
	}
.mg {
	grid-column: 2 / 3 ;
	grid-row: 4 ;
	}
.mh {
	grid-column: 1 / 3 ;
	grid-row: 5 ;
	}
.mi {
	grid-column: 1 / 3 ;
	grid-row: 7 ;
	}
.mj {
	grid-column: 1 / 2 ;
	grid-row: 8 ;
	}
.mk {
	grid-column: 1 / 2 ;
	grid-row: 9 ;
	}
.ml {
	grid-column: 1 / 2 ;
	grid-row: 10 ;
	}
.mm {
	grid-column: 2 / 3 ;
	grid-row: 8 ;
	}
.mn {
	grid-column: 2 / 3 ;
	grid-row: 9 ;
	}
.mo {
	grid-column: 2 / 3 ;
	grid-row: 10 ;
	}
.mp {
	grid-column: 1 / 3 ;
	grid-row: 6 ;
	}
.mq {
	grid-column: 1 / 3 ;
	grid-row: 11;
	}
.mr {
	grid-column: 1 / 3 ;
	grid-row: 12 ;
	}
.mt {
	grid-column: 1 / 3 ;
	grid-row: 13 ;
	}
.mu {
	grid-column: 1 / 3 ;
	grid-row: 14 ;
	}
}

